from datetime import datetime
from io       import TextIOWrapper
from sys      import stdout

LOG_SILENT  = 0
LOG_NORMAL  = 1
LOG_VERBOSE = 2

class Logger():

    def __init__(self, log_level=LOG_NORMAL, file=stdout, flush_immediately=False, close_on_exit=False):
        self.log_level = log_level
        self.file = file
        self.flush_immediately = flush_immediately
        self.close_on_exit = close_on_exit

    def __enter__(self):
        if isinstance(self.file, str):
            # treat as filename
            self.__file = open(self.file, 'w')

        elif isinstance(self.file, TextIOWrapper):
            # treat as file (e.g. for stdout)
            self.__file = self.file

        return self

    def __exit__(self, type, value, traceback):
        if self.close_on_exit:
            self.__file.close()

    def log(self, *msg, log_level=LOG_NORMAL, id=0, **kwargs):
        if log_level <= self.log_level and not self.__file.closed:
            timestamp = datetime.now().strftime("[%H:%M:%S.%f]")
            print(timestamp, '('+str(id)+')', *msg, **kwargs, file=self.__file, flush=self.flush_immediately)


import numpy as np


class Evaluator:
    """Evaluator.

    Base class that evaluates individuals. Must be implemented
    to match Individual implementation.
    """

    def __init__(self, logger=None):
        self.logger_ = logger

    def log(self, *args, log_level=1, id=0, **kwargs):
        if self.logger_ is not None:
            self.logger_.log(*args, log_level=log_level, id=id, **kwargs)

    def eval(self, individual):
        """eval.

        Evaluate individual, returning a numerical loss value.

        individual: Individual to evaluate.
        """
        raise NotImplementedError()

    def eval_all(self, individuals):
        """eval_all.

        Evaluate individuals in a given iterable, return list of loss values.

        individuals: iterable containing Individuals.
        """
        losses = []
        for ind in individuals:
            if ind.loss_ is None:
                loss = self.eval(ind)
                ind.loss_ = loss
            losses.append(ind.loss_)
        return losses



import numpy as np

from .individual import Individual
from .evaluator import Evaluator
from .optimizer import Optimizer
from .logger import Logger, LOG_VERBOSE

N = 10 # size of vector


class RealVector(Individual):

    def __init__(self, vec, logger=None):
        super().__init__(logger)
        self.vec = vec

    def __repr__(self):
        return str(self.vec) + " (loss={})".format(self.loss_)

    def copy(self):
        ind = self.__class__(self.vec.copy(), self.logger_)
        return ind

    @classmethod
    def random(cls, size, logger=None):
        return cls(np.random.normal(loc=0, scale=2, size=N), logger=logger)

    def mutate(self, mutation_rate=0.1, mutation_width=1):       
        if isinstance(mutation_width, int):
            # interpret as absolute number
            k = mutation_width
        elif isinstance(mutation_width, float):
            # interpret as probability per vector element
            k = max(1, np.random.binomial(self.vec.size, mutation_width))

        indices = np.random.choice(self.vec.size, size=k, replace=False)
        noise = np.random.normal(loc=0, scale=mutation_rate, size=k)
        self.vec[indices] += noise
        self.log('Added ' + ', '.join(['{} to vec[{}]'.format(noise_, ix) for ix, noise_ in zip(indices, noise)]) + 'of #'+str(self.id_), log_level=2, id=12)

    def cross(self, other, crossover_rate=0.5):
        # uniform crossover
        k = max(1, np.random.binomial(self.vec.size, crossover_rate))
        indices = np.random.choice(self.vec.size, size=k, replace=False)
        self.vec[indices] = other.vec[indices]
        self.log('#{} took indices {} from #{}'.format(self.id_, ', '.join(map(str,indices)), other.id_), log_level=2, id=13)


class RealVectorEvaluator(Evaluator):

    def __init__(self, target_vector, logger=None):
        super().__init__(logger)
        self.target_vec = target_vector

    def eval(self, ind):
        return np.linalg.norm(ind.vec - self.target_vec, 2)


def print_lines(ls):
    for l in ls:
        print(l)

conf = {
    'parents': 5,
    'offspring': 10,

    'do_crossover': True,
    'do_self_adaption': True,

    'mutation_prob': 0.9,
    'crossover_kwargs': {
        'crossover_rate': 0.5},
    'mutation_kwargs': {
        'mutation_rate': 1,
        'mutation_width': 0.1},

    'mutation_kwargs_bounds': {
        'mutation_width': (0.0, 1.0)
    }
}


with Logger(log_level=LOG_VERBOSE, file="main_log.txt") as l:
    ev = RealVectorEvaluator(np.random.uniform(-20, 20, size=N), logger=l)
    print('target: {}'.format(ev.target_vec))
    opt = Optimizer(ev, config=conf, logger=l)
    res = opt.run([RealVector.random(N, logger=l) for _ in range(5)], generations=50)

#print_lines(res)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(name='eao',
      version='0.6',
      description='Evolutionary Optimization',
      long_description=long_description,
      long_description_content_type='text/markdown',
      keywords=['evolution', 'evolutionary', 'algorithm', 'optimization', 'genetic'],
      author=u'Sascha M\u00fccke',
      author_email='sascha.muecke@tu-dortmund.de',
      url='https://github.com/smuecke/eao',
      license='GNU GPLv3',
      packages=setuptools.find_packages(),
      python_requires='>=3.5',
      install_requires=['numpy', 'tqdm'])
